#!/usr/bin/env python
# coding=utf-8

import argparse
import logging
import math
import os
import random
import datasets
import torch
from functools import partial
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import set_seed
from datasets import load_dataset
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

import transformers
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    LlamaTokenizer,
    LlamaTokenizerFast,
    SchedulerType,
    DataCollatorForSeq2Seq,
    get_scheduler,
    GPTNeoXTokenizerFast,
    GPT2Tokenizer,
    OPTForCausalLM,
    BitsAndBytesConfig,
)
from peft import LoraConfig, TaskType, get_peft_model
from peft import prepare_model_for_kbit_training

logger = get_logger(__name__)

import json
import numpy as np
import re
import os
import csv
import yaml
from datetime import datetime
from collections import defaultdict
from pathlib import Path
from Levenshtein import ratio
from math import ceil
import attr
from datasets import load_from_disk

WS_PATH = Path(os.environ["WS_PATH"]) 

HP_CONFIGS = {
    "bloke_hp": {
        "use_lora": True,
        "lora_rank": 16,
        "lora_alpha": 16,
        "lora_dropout": 0.05,
        "lora_modules": ["q_proj", "k_proj", "v_proj", "o_proj"],
        "learning_rate": 3e-4,
        "warmup_ratio": 0.03,
        "lr_scheduler_type": SchedulerType.LINEAR,
        "weight_decay": 0.,
    },
    "oi_lora_hp": {
        "use_lora": True,
        "lora_rank": 256,
        "lora_alpha": 256,
        "lora_dropout": 0.1,
        "lora_modules": ["q_proj", "o_proj", "v_proj", "k_proj", "gate_proj", "up_proj", "down_proj""q_proj", "k_proj", "v_proj", "o_proj"],
        "learning_rate": 1e-4,
        "warmup_ratio": 0.03,
        "lr_scheduler_type": SchedulerType.LINEAR,
        "weight_decay": 0.,
    },
    "oi_qlora_hp": {
        "use_lora": True,
        "use_qlora": True,
        "lora_rank": 64,
        "lora_alpha": 64,
        "lora_dropout": 0.1,
        "lora_modules": ["q_proj", "o_proj", "v_proj", "k_proj", "gate_proj", "up_proj", "down_proj""q_proj", "k_proj", "v_proj", "o_proj"],
        "learning_rate": 2e-4,
        "warmup_ratio": 0.03,
        "lr_scheduler_type": SchedulerType.LINEAR,
        "weight_decay": 0.,
    },
    "oi_no_lora_hp": {
        "learning_rate": 2e-5,
        "warmup_ratio": 0.03,
        "lr_scheduler_type": SchedulerType.LINEAR,
        "weight_decay": 0.,
    }
}

SS_CONFIGS = {
    "ss_base": {},
    "ss_v1": {
        "ss_stratified_new_samples": True,
        "ss_dropout": True,
    },
    "ss_v2": {
        "ss_stratified_new_samples": True,
        "ss_dropout": True,
        "ss_competency_drop": True,
    }
}
ratio_configs = {}
for k, v in SS_CONFIGS.items():
    ratio_configs[f"{k}_r1"] = {
        **v,
        "ss_warmup_up_ratio": 0.25,
        "ss_new_samples_ratio": 0.15,
    },
SS_CONFIGS.update(ratio_configs)

DS_MIXES = {
    "all": [
                "super_ni",
                "cot",
                "flan_v2",
                "self_instruct",
                "unnatural_instructions",
                "stanford_alpaca",
                "dolly",
                "oasst1",
                "code_alpaca",
                "gpt4_alpaca",
                "baize",
                "sharegpt",
                "lima",
                "wizardlm",
                "open_orca",
            ],
    "tulu_v1_human_mix": [
                "flan_v2",
                "cot",
                "dolly",
                "oasst1",
    ],
    "tulu_v1_gpt_mix": [
                "gpt4_alpaca",
                "code_alpaca",
                "sharegpt",
    ],
    "tulu_v1_mix": [
                "flan_v2",
                "cot",
                "dolly",
                "oasst1",
                "gpt4_alpaca",
                "code_alpaca",
                "sharegpt",
    ],
    "tulu_v2_human_mix": [
                "flan_v2",
                "cot",
                "lima",
                "oasst1",
    ],
    "tulu_v2_gpt_mix": [
                "code_alpaca",
                "sharegpt",
                "wizardlm",
                "open_orca",
    ],
    "tulu_v2_mix": [
                "flan_v2",
                "cot",
                "lima",
                "oasst1",
                "code_alpaca",
                "sharegpt",
                "wizardlm",
                "open_orca",
    ],
    "reasoning_v1": [
        "math_prompt",
        "gsm_prompt"
    ],
    "reasoning_v2": [
        "math_prompt_2",
        "gsm_prompt_2"
    ]
        
}


############################################
@attr.s
class CustomSuperMemo2:
    easiness = attr.ib(validator=attr.validators.instance_of(float))
    interval = attr.ib(validator=attr.validators.instance_of(int))
    successful_repetitions = attr.ib(validator=attr.validators.instance_of(int))
    total_repetitions = attr.ib(validator=attr.validators.instance_of(int))

    @staticmethod
    def first_review(
        quality: int,
    ) -> "CustomSuperMemo2":
        return CustomSuperMemo2(2.5, 0, 1, 1).review(quality)

    def review(
        self,
        quality: int,
    ) -> "CustomSuperMemo2":
        self.total_repetitions += 1
        if quality < 3:
            self.interval = 1
            self.successful_repetitions = 0
        else:
            if self.successful_repetitions == 0:
                self.interval = 1
            elif self.successful_repetitions == 1:
                self.interval = 6
            else:
                self.interval = ceil(self.interval * self.easiness)

            self.successful_repetitions = self.successful_repetitions + 1

        self.easiness += 0.1 - (5 - quality) * (0.08 + (5 - quality) * 0.02)
        if self.easiness < 1.3:
            self.easiness = 1.3

        return self

def add_custom_args(parser):
    parser.add_argument(
        "--dry_run",
        action="store_true",
    )
    parser.add_argument(
        "--base_data_dir",
        type=str,
        default=os.path.join(WS_PATH, "data/train"),
    )
    parser.add_argument(
        "--base_output_dir",
        type=str,
        default=os.path.join(WS_PATH, "exp_artifacts"),
    )
    parser.add_argument(
        "--base_model_dir",
        type=str,
        default=os.path.join(WS_PATH, "models"),
    )
    
    parser.add_argument(
        "--num_gpus",
        type=int,
    )
    
    parser.add_argument(
        "--exp_name",
        type=str,
        default=None,
    )
    
    parser.add_argument(
        "--config_file",
        default=None,
    )
    parser.add_argument(
        "--train_ds",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--mix_size",
        type=str,
        default=None
    )
    parser.add_argument(
        "--per_device_eval_batch_size",
        type=int,
        default=8,
    )
    parser.add_argument(
        "--lora_modules",
        type=list,
        default=None,
    )
    parser.add_argument(
        "--hp_config",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--total_train_batch_size",
        type=int,
        default=None
    )
    parser.add_argument(
        "--mixture_name",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--tokenized_ds",
        type=str,
        default=None
    )
    
    parser.add_argument(
        "--debug",
        action="store_true",
    )
    
    parser.add_argument(
        "--ss_config",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--use_spaced_sampling",
        action="store_true",
    )

    parser.add_argument(
        "--ss_dropout",
        action="store_true",
    )
    parser.add_argument(
        "--ss_skip_prev_samples",
        action="store_true",
    )
    parser.add_argument(
        "--ss_warmup_ratio",
        type=float,
        default=0.1,
    )
    parser.add_argument(
        "--ss_new_sample_ratio",
        type=float,
        default=0.05,
    )
    parser.add_argument(
        "--ss_stratified_new_samples",
        action="store_true",
    )
    parser.add_argument(
        "--ss_competency_drop",
        action="store_true",
    )
    
    return parser

def get_model_name_from_path(path):
    # For shorter names
    mapping = {
        'meta_llama_Llama_2_7b_hf': "llama2_7b",
        'meta_llama_Llama_2_13b_hf': "llama2_13b"
    }
    return mapping.get(path.name, path.name)

def create_experiment_name(args):
    if args.mixture_name:
        train_dataset_name = args.mixture_name
    else:
        train_dataset_name = Path(args.train_file).parent.name

    name_parts = [args.ss_config] if args.use_spaced_sampling else []
    name_parts.append(train_dataset_name)
    if args.mix_size is not None:
        name_parts.append(args.mix_size)
    name_parts.extend([get_model_name_from_path(Path(args.model_name_or_path))])
    
    if args.hp_config is not None:
        name_parts.append(args.hp_config)
        
    current_file_parent_folder = Path(__file__).parent.name
    contains_time_stamp = re.match(r"^.*_(\d{14})$", current_file_parent_folder)
    if not contains_time_stamp:
        current_time = datetime.now().strftime("%Y%m%d%H%M%S")
    else:
        current_time = contains_time_stamp.group(1)

    name_parts.append(current_time)
    
    return "_".join(name_parts)
    


def set_args_from_config_file(args):
    if args.config_file is not None:
        # load config file
        with open(args.config_file, 'r') as file:
            yaml_config = yaml.safe_load(file)
 
        for key, value in yaml_config.items():
            if not hasattr(args, key):
                raise ValueError(f"Invalid config key: {key}. Does not exist in args")
            
            # Check if value is int or float
            if type(getattr(args, key)) in [int, float]:
                value = type(getattr(args, key))(value)

            setattr(args, key, value)
            
    if args.debug:
        print("((((((((((((((((((((((((DEBUG))))))))))))))))))))))))")
        # debug
        # args.model_name_or_path = "opt_350m"
        # args.model_name_or_path = "facebook_opt_350m"
        # args.model_name_or_path = "falcon_7b"
        # args.model_name_or_path = "mpt_7b"
        args.preprocessing_num_workers = 32
        args.with_tracking = False
        args.report_to = None
    else:
        args.with_tracking = True
        args.report_to = "wandb"

    # Expand some args
    if args.mixture_name is not None:
        tokenized_ds_candidate = f'{args.mixture_name}_{args.mix_size}_llama2'
        tokenized_ds_candidate_path = Path(args.base_data_dir) / tokenized_ds_candidate
        if tokenized_ds_candidate_path.exists():
            args.tokenized_ds = tokenized_ds_candidate_path.as_posix()
        elif args.mixture_name in DS_MIXES:
            all_ds_names = DS_MIXES[args.mixture_name]
            train_files = []
            for ds_name in all_ds_names:
                sized_ds_name = f"{ds_name}_{args.mix_size}" if args.mix_size is not None else ds_name
                train_files.append((Path(args.base_data_dir) / ds_name /f"{sized_ds_name}_data.jsonl").as_posix())
            args.train_file = train_files
        else:
            # Single dataset
            sized_ds_name = f"{args.mixture_name}_{args.mix_size}" if args.mix_size is not None else args.mixture_name 
            args.train_file = (Path(args.base_data_dir) / args.mixture_name /f"{args.mixture_name}_data.jsonl").as_posix()
            
    if args.train_file is None and args.tokenized_ds is None:
        raise ValueError("train data is None")
    
    if args.ss_config is not None:
        args.use_spaced_sampling = True
        for key, value in SS_CONFIGS[args.ss_config].items():
            if not hasattr(args, key):
                raise ValueError(f"Invalid config key: {key}. Does not exist in args")
            setattr(args, key, value)
        
    if not Path(args.model_name_or_path).exists() and '70b' not in args.model_name_or_path:
        args.model_name_or_path = (Path(args.base_model_dir) / args.model_name_or_path).as_posix()
        
    if args.exp_name is None:
        args.exp_name = create_experiment_name(args)

    if args.output_dir is None:
        args.output_dir = (Path(args.base_output_dir) / args.exp_name).as_posix()
        
    if args.hp_config is not None:
        for key, value in HP_CONFIGS[args.hp_config].items(): 
            if not hasattr(args, key):
                raise ValueError(f"Invalid config key: {key}. Does not exist in args")
            if key == 'lora_modules' and 'llama2' not in args.model_name_or_path:
                if 'falcon' in args.model_name_or_path or 'mpt' in args.model_name_or_path:
                    value = ['query_key_value']
                else:
                    continue
            setattr(args, key, value)

    # Some defaults
    
    args.gradient_accumulation_steps = args.total_train_batch_size // args.num_gpus // args.per_device_train_batch_size
        
        
    os.environ["WANDB_NAME"] = args.exp_name
    args.use_flash_attn = True
    if 'mpt' in args.model_name_or_path:
        args.use_slow_tokenizer = False
    else: 
        args.use_slow_tokenizer = True

    return args

def add_id(example, idx):
    example["id"] = idx
    return example

def compute_tokens_edit_score(tokenizer, pred, label, label_mask, reasoning=True):

    start_label_idx = torch.where(label_mask != False)[0][0].item()

    start_pred_idx = start_label_idx - 1 # Handle the shifting done during training 
    try:
        end_pred_idx = torch.where(pred[start_pred_idx:] == tokenizer.eos_token_id)[0][0] + start_pred_idx 
    except:
        # Try bos
        try:
            end_pred_idx = torch.where(pred[start_pred_idx:] == tokenizer.bos_token_id)[0][-1] + start_pred_idx 
        except:
            end_pred_idx = len(pred) - 1

    pred_mask = torch.zeros_like(pred, dtype=torch.bool)
    pred_mask[start_pred_idx:end_pred_idx] = True

    curr_label = torch.masked_select(label, label_mask)
    curr_pred = torch.masked_select(pred, pred_mask)

    curr_label_list = curr_label.tolist()
    curr_pred_list = curr_pred.tolist()

    if True:
        # For logging
        curr_pred_str = tokenizer.decode(curr_pred_list, skip_special_tokens=True)
        curr_label_str = tokenizer.decode(curr_label_list, skip_special_tokens=True)
    else:
        curr_pred_str = None
        curr_label_str = None
        
        
    ss = None
    final_answer_str = "final answer:"
    if reasoning:
        if  final_answer_str not in curr_label_str or final_answer_str not in curr_pred_str:
            ss = 0
    
    if ss is None:
        sr = ratio(curr_label_list, curr_pred_list)
        ss = int(sr * 5)

    if False:
        print("-----")
        print("Prediction: \n" + curr_pred_str)
        print("Label: \n" + curr_label_str)
        print("Score: " + str(ss))
        print("-----")

    return ss, curr_pred_str, curr_label_str


def evaluate_batch(model, batch, tokenizer):
    sample_ids = []
    scores = []
    pred_str_list = []
    label_str_list = []

    with torch.no_grad():
        outputs = model(input_ids=batch['input_ids'], labels=batch['labels'], attention_mask=batch['attention_mask'], use_cache=False)

    # Greedy decode
    preds = torch.argmax(outputs.logits, -1)

    labels = batch["labels"]
    label_masks = labels != -100

    for i in range(0, len(labels)):
        label = labels[i]
        label_mask = label_masks[i]
        pred = preds[i]

        score, pred_str, label_str = compute_tokens_edit_score(tokenizer, pred, label, label_mask)

        pred_str_list.append(pred_str)
        label_str_list.append(label_str)

        scores.append(score)
        sample_ids.append(batch['id'][i].item())

    return (
        sample_ids,
        scores,
        pred_str_list,
        label_str_list,
    )


def write_review_log(output_dir, review_log, iter_count, accelerator):
    log_folder = Path(output_dir) / "ss_logs" 
    log_folder.mkdir(exist_ok=True, parents=True)
    output_file = log_folder / f"review_log_p{accelerator.process_index}_i{iter_count}.csv"
    with output_file.open("w", newline="") as f:
        writer =csv.DictWriter(f, fieldnames=review_log[0].keys())
        writer.writeheader()
        for row in review_log:
            writer.writerow(row)
            
            
def get_category_indices(dataset):
    category_indices = defaultdict(list)
    for sample in dataset:
        category_indices[sample['category_name']].append(sample['id'])

    return category_indices

def ss_set_train_dataset(lm_datasets, args, accelerator):
    # SpaceSampling: Add ID to each example
    ds = lm_datasets['train'].train_test_split(test_size=0.1, seed=args.seed)

    # Shuffle the dataset, to keep the id = idx
    ds = ds.shuffle(seed=args.seed)

    with accelerator.main_process_first():
        ds['train'] = ds['train'].map(
            add_id,
            batched=False,
            num_proc=args.preprocessing_num_workers,
            load_from_cache_file=not args.overwrite_cache,
            with_indices=True,
            desc="Adding id to instruction data",
        )

    with accelerator.main_process_first():
        ds['test'] = ds['test'].map(
            add_id,
            batched=False,
            num_proc=args.preprocessing_num_workers,
            load_from_cache_file=not args.overwrite_cache,
            with_indices=True,
            desc="Adding id to instruction data",
        )

    # For qualitative analysis needs
    sample_metadata_map = defaultdict(dict)
    category_ids = defaultdict(list)
    for split in ds:
        for sample in ds[split]:
            if sample['id'] in sample_metadata_map:
                raise ValueError(f"Duplicate sample id: {sample['id']}")

            sample_metadata_map[split][sample['id'].item()] = {'sample_uuid': sample['sample_uuid'], 'category_name': sample['category_name']}
            if split == 'train':
                category_ids[sample['category_name']].append(sample['id'].item())

    return ds, sample_metadata_map, category_ids

def load_tokenized_dataset(tokenized_ds_path):
    print(f"Loading pretokenized data at {tokenized_ds_path}")
    ds = load_from_disk(Path(tokenized_ds_path) / "tokenized_dataset")

    parent_folder = Path(tokenized_ds_path)
    category_ids_file = parent_folder /  'category_indices.json'
    with category_ids_file.open('r') as f:
        category_ids = json.load(f)

    sample_metadata_map_file = parent_folder / 'sample_metadata_map.json'
    with sample_metadata_map_file.open('r') as f:
        sample_metadata_map = json.load(f)

    # Sanity
    ds_train = ds['train']
    random_ids = random.choices(range(len(ds_train)), k=3)
    for random_id in random_ids:
        assert ds_train[random_id]['sample_uuid'] == sample_metadata_map['train'][str(random_id)]['sample_uuid']

    return ds, sample_metadata_map, category_ids

def store_tokenized_ds(args, train_dataset, sample_metadata_map, category_indices):
    model_name = Path(args.model_name_or_path).name
    dataset_parent_path = Path(args.base_data_dir) / f"{args.mixture_name}_{args.mix_size}_{model_name.split('_')[0]}"
    dataset_parent_path.mkdir(exist_ok=True)

    dataset_path = dataset_parent_path / 'tokenized_dataset'
    train_dataset.save_to_disk(dataset_path.as_posix())

    category_ids_file = dataset_parent_path /  'category_indices.json'
    with category_ids_file.open('w') as f:
        json.dump(category_indices, f)

    sample_metadata_map_file = dataset_parent_path / 'sample_metadata_map.json'
    with sample_metadata_map_file.open('w') as f:
        json.dump(sample_metadata_map, f)

    print(f'Wrote tokenized ds to {dataset_parent_path}')
    
    return dataset_parent_path

    
#############################################


def parse_args():
    parser = argparse.ArgumentParser(description="Finetune a transformers model on a causal language modeling task")
    parser.add_argument(
        "--dataset_name",
        type=str,
        default=None,
        help="The name of the dataset to use (via the datasets library).",
    )
    parser.add_argument(
        "--dataset_config_name",
        type=str,
        default=None,
        help="The configuration name of the dataset to use (via the datasets library).",
    )
    parser.add_argument(
        "--train_file", type=str, default=None, help="A csv or a json file containing the training data."
    )
    parser.add_argument(
        "--model_name_or_path",
        type=str,
        help="Path to pretrained model or model identifier from huggingface.co/models.",
        required=False,
    )
    parser.add_argument(
        "--config_name",
        type=str,
        default=None,
        help="Pretrained config name or path if not the same as model_name",
    )
    parser.add_argument(
        "--use_lora",
        action="store_true",
        help="If passed, will use LORA (low-rank parameter-efficient training) to train the model.",
    )
    parser.add_argument(
        "--lora_rank",
        type=int,
        default=64,
        help="The rank of lora.",
    )
    parser.add_argument(
        "--lora_alpha",
        type=float,
        default=16,
        help="The alpha parameter of lora.",
    )
    parser.add_argument(
        "--lora_dropout",
        type=float,
        default=0.1,
        help="The dropout rate of lora modules.",
    )
    parser.add_argument(
        "--use_flash_attn",
        action="store_true",
        help="If passed, will use flash attention to train the model.",
    )
    parser.add_argument(
        "--tokenizer_name",
        type=str,
        default=None,
        help="Pretrained tokenizer name or path if not the same as model_name",
    )
    parser.add_argument(
        "--use_slow_tokenizer",
        action="store_true",
        help="If passed, will use a slow tokenizer (not backed by the 🤗 Tokenizers library).",
    )
    parser.add_argument(
        "--max_seq_length",
        type=int,
        default=512,
        help="The maximum total sequence length (prompt+completion) of each training example.",
    )
    parser.add_argument(
        "--per_device_train_batch_size",
        type=int,
        default=8,
        help="Batch size (per device) for the training dataloader.",
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        default=5e-5,
        help="Initial learning rate (after the potential warmup period) to use.",
    )
    parser.add_argument("--weight_decay", type=float, default=0.0, help="Weight decay to use.")
    parser.add_argument("--num_train_epochs", type=int, default=3, help="Total number of training epochs to perform.")
    parser.add_argument(
        "--max_train_steps",
        type=int,
        default=None,
        help="Total number of training steps to perform. If provided, overrides num_train_epochs.",
    )
    parser.add_argument(
        "--gradient_accumulation_steps",
        type=int,
        default=1,
        help="Number of updates steps to accumulate before performing a backward/update pass.",
    )
    parser.add_argument(
        "--lr_scheduler_type",
        type=SchedulerType,
        default="linear",
        help="The scheduler type to use.",
        choices=["linear", "cosine", "cosine_with_restarts", "polynomial", "constant", "constant_with_warmup"],
    )
    parser.add_argument(
        "--warmup_ratio", type=float, default=0, help="Ratio of total training steps used for warmup."
    )
    parser.add_argument("--output_dir", type=str, default=None, help="Where to store the final model.")
    parser.add_argument("--seed", type=int, default=None, help="A seed for reproducible training.")
    parser.add_argument(
        "--preprocessing_num_workers",
        type=int,
        default=None,
        help="The number of processes to use for the preprocessing.",
    )
    parser.add_argument(
        "--overwrite_cache", action="store_true", help="Overwrite the cached training and evaluation sets"
    )
    parser.add_argument(
        "--checkpointing_steps",
        type=str,
        default=None,
        help="Whether the various states should be saved at the end of every n steps, or 'epoch' for each epoch.",
    )
    parser.add_argument(
        "--logging_steps",
        type=int,
        default=None,
        help="Log the training loss and learning rate every logging_steps steps.",
    )
    parser.add_argument(
        "--resume_from_checkpoint",
        type=str,
        default=None,
        help="If the training should continue from a checkpoint folder.",
    )
    parser.add_argument(
        "--with_tracking",
        action="store_true",
        help="Whether to enable experiment trackers for logging.",
    )
    parser.add_argument(
        "--report_to",
        type=str,
        default="all",
        help=(
            'The integration to report the results and logs to. Supported platforms are `"tensorboard"`,'
            ' `"wandb"`, `"comet_ml"` and `"clearml"`. Use `"all"` (default) to report to all integrations.'
            "Only applicable when `--with_tracking` is passed."
        ),
    )
    parser.add_argument(
        "--low_cpu_mem_usage",
        action="store_true",
        help=(
            "It is an option to create the model as an empty shell, then only materialize its parameters when the pretrained weights are loaded."
            "If passed, LLM loading time and RAM consumption will be benefited."
        ),
    )
    parser.add_argument(
        "--gradient_checkpointing",
        action="store_true",
        help=(
            "Turn on gradient checkpointing. Saves memory but slows training."
        ),
    )
    parser.add_argument(
        "--use_qlora",
        action="store_true",
        help=(
            "Use qLoRA training - main thing is initialising model in quantised form. Not compatible with deepspeed."
        ),
    )
    parser.add_argument(
        '--clip_grad_norm',
        type=float,
        default=-1,
        help='Clip gradient norm. Not compatible with deepspeed (use deepspeed config instead).',
    )
    parser.add_argument(
        '--use_8bit_optimizer',
        action='store_true',
        help='Use 8bit optimizer from bitsandbytes. Not compatible with deepspeed (use deepspeed config instead).',
    )
    
    parser = add_custom_args(parser)

    args = parser.parse_args()
    
    args = set_args_from_config_file(args)

    return args


def encode_with_prompt_completion_format(example, tokenizer, max_seq_length):
    '''
    Here we assume each example has 'prompt' and 'completion' fields.
    We concatenate prompt and completion and tokenize them together because otherwise prompt will be padded/trancated 
    and it doesn't make sense to follow directly with the completion.
    '''
    # if prompt doesn't end with space and completion doesn't start with space, add space
    if not example['prompt'].endswith((' ', '\n', '\t')) and not example['completion'].startswith((' ', '\n', '\t')):
        example_text = example['prompt'] + ' ' + example['completion']
    else:
        example_text = example['prompt'] + example['completion']
    example_text = example_text + tokenizer.eos_token
    tokenized_example = tokenizer(example_text, return_tensors='pt', max_length=max_seq_length, truncation=True)
    input_ids = tokenized_example.input_ids
    labels = input_ids.clone()
    tokenized_prompt = tokenizer(example['prompt'], return_tensors='pt', max_length=max_seq_length, truncation=True)
    # mask the prompt part for avoiding loss
    labels[:, :tokenized_prompt.input_ids.shape[1]] = -100
    attention_mask = torch.ones_like(input_ids)
    return {
        'input_ids': input_ids.flatten(),
        'labels': labels.flatten(),
        'attention_mask': attention_mask.flatten(),
        'sample_uuid': example['id'],
        'category_name': example['category_name'],

    }


def encode_with_messages_format(example, tokenizer, max_seq_length):
    '''
    Here we assume each example has a 'messages' field Each message is a dict with 'role' and 'content' fields.
    We concatenate all messages with the roles as delimiters and tokenize them together.
    '''
    messages = example['messages']
    if len(messages) == 0:
        raise ValueError('messages field is empty.')
    
    def _concat_messages(messages):
        message_text = ""
        for message in messages:
            if message["role"] == "system":
                message_text += "<|system|>\n" + message["content"].strip() + "\n"
            elif message["role"] == "user":
                message_text += "<|user|>\n" + message["content"].strip() + "\n"
            elif message["role"] == "assistant":
                message_text += "<|assistant|>\n" + message["content"].strip() + tokenizer.eos_token + "\n"
            else:
                raise ValueError("Invalid role: {}".format(message["role"]))
        return message_text
        
    example_text = _concat_messages(messages).strip()
    tokenized_example = tokenizer(example_text, return_tensors='pt', max_length=max_seq_length, truncation=True)
    input_ids = tokenized_example.input_ids
    labels = input_ids.clone()

    # mask the non-assistant part for avoiding loss
    for message_idx, message in enumerate(messages):
        if message["role"] != "assistant":
            if message_idx == 0:
                message_start_idx = 0
            else:
                message_start_idx = tokenizer(
                    _concat_messages(messages[:message_idx]), return_tensors='pt', max_length=max_seq_length, truncation=True
                ).input_ids.shape[1]
            if message_idx < len(messages) - 1 and messages[message_idx+1]["role"] == "assistant":
                # here we also ignore the role of the assistant
                messages_so_far = _concat_messages(messages[:message_idx+1]) + "<|assistant|>\n"
            else:
                messages_so_far = _concat_messages(messages[:message_idx+1])
            message_end_idx = tokenizer(
                messages_so_far,
                return_tensors='pt', 
                max_length=max_seq_length, 
                truncation=True
            ).input_ids.shape[1]
            labels[:, message_start_idx:message_end_idx] = -100
            
            if message_end_idx >= max_seq_length:
                break

    attention_mask = torch.ones_like(input_ids)
    return {
        'input_ids': input_ids.flatten(),
        'labels': labels.flatten(),
        'attention_mask': attention_mask.flatten(),
        'sample_uuid': example['id'],
        'category_name': example['category_name'],
    }


def save_with_accelerate(accelerator, model, tokenizer, output_dir, args):
    unwrapped_model = accelerator.unwrap_model(model)
    # When doing multi-gpu training, we need to use accelerator.get_state_dict(model) to get the state_dict.
    # Otherwise, sometimes the model will be saved with only part of the parameters.
    # Also, accelerator needs to use the wrapped model to get the state_dict.
    state_dict = accelerator.get_state_dict(model)
    if args.use_lora:
        # When using lora, the unwrapped model is a PeftModel, which doesn't support the is_main_process 
        # and has its own save_pretrained function for only saving lora modules.
        # We have to manually specify the is_main_process outside the save_pretrained function.
        if accelerator.is_main_process:
            unwrapped_model.save_pretrained(output_dir, state_dict=state_dict)
    else:
        unwrapped_model.save_pretrained(
            output_dir, is_main_process=accelerator.is_main_process, save_function=accelerator.save, state_dict=state_dict
        )
        
from cl_llm.eval.eval_gsm import extract_answer_number    
from cl_llm.eval.eval_math import process_results
from cl_llm.eval.utils import last_boxed_only_string

def evaluate_training(args, accelerator, model, tokenizer, eval_dataloader, sample_metadata, completed_steps):
    model.eval()
    gsm_res = []
    math_res = []
    total_eval_loss = 0.0
    for step, batch in enumerate(tqdm(eval_dataloader, disable=not accelerator.is_local_main_process, desc="Train Eval")):
        with torch.no_grad():
            outputs = model(input_ids=batch['input_ids'], labels=batch['labels'], attention_mask=batch['attention_mask'], use_cache=False)
            loss = outputs.loss
            generated_tokens = torch.argmax(outputs.logits, -1)
            generated_tokens = accelerator.pad_across_processes(
                generated_tokens, dim=1, pad_index=tokenizer.pad_token_id
            )
            labels = batch["labels"]

            # If we did not pad to max length, we need to pad the labels too
            labels = accelerator.pad_across_processes(batch["labels"], dim=1, pad_index=tokenizer.pad_token_id)

            generated_tokens, labels, loss = accelerator.gather_for_metrics((generated_tokens, labels, loss))
            total_eval_loss += loss.mean().item()
            generated_tokens = generated_tokens.cpu().numpy()
            labels = labels.cpu().numpy()

            # Replace -100 in the labels as we can't decode them.
            labels = np.where(labels != -100, labels, tokenizer.pad_token_id)

            if isinstance(generated_tokens, tuple):
                generated_tokens = generated_tokens[0]

            decoded_preds = tokenizer.batch_decode(generated_tokens, skip_special_tokens=True)
            decoded_labels = tokenizer.batch_decode(labels, skip_special_tokens=True)
            
        for id, pred, label in zip(batch['id'], decoded_preds, decoded_labels):
            category = sample_metadata['test'][str(id.item())]['category_name']
            if 'gsm' in category:
                pred_final_answer = extract_answer_number(pred)
                label_final_answer = extract_answer_number(label)
                assert label_final_answer is not None
                if pred_final_answer != None:
                    gsm_res.append(float(pred_final_answer) == float(label_final_answer))
                else:
                    gsm_res.append(False)
            elif 'math' in category:
                label_final_answer = last_boxed_only_string(label)
                valid = process_results(None, pred, label_final_answer)
                math_res.append(valid)
            else:
                raise ValueError(f"Invalid category: {category}")


    results = {
        "eval/loss": total_eval_loss / len(eval_dataloader), 
        "eval/gsm_acc": sum(gsm_res) / len(gsm_res),
        "eval/math_acc": sum(math_res) / len(math_res),
    }
    
    print('Eval results: ', results)
    if args.with_tracking:
        accelerator.log(results, step=completed_steps)
            
    model.train()

def main():
    args = parse_args()
    if args.dry_run:
        print(f'EXP_NAME={args.exp_name}')
        return

    # A hacky way to make llama work with flash attention
    if args.use_flash_attn:
        from llama_flash_attn_monkey_patch import replace_llama_attn_with_flash_attn
        replace_llama_attn_with_flash_attn()

    # Initialize the accelerator. We will let the accelerator handle device placement for us in this example.
    # If we're using tracking, we also need to initialize it here and it will by default pick up all supported trackers
    # in the environment
    accelerator_log_kwargs = {}

    if args.with_tracking:
        accelerator_log_kwargs["log_with"] = args.report_to
        accelerator_log_kwargs["project_dir"] = args.output_dir

    accelerator = Accelerator(gradient_accumulation_steps=args.gradient_accumulation_steps, **accelerator_log_kwargs, mixed_precision="bf16")
    # Make one log on every process with the configuration for debugging.
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.info(accelerator.state, main_process_only=False)
    if accelerator.is_local_main_process:
        datasets.utils.logging.set_verbosity_warning()
        transformers.utils.logging.set_verbosity_info()
    else:
        datasets.utils.logging.set_verbosity_error()
        transformers.utils.logging.set_verbosity_error()

    # If passed along, set the training seed now.
    if args.seed is not None:
        set_seed(args.seed)

    if accelerator.is_main_process:
        if args.output_dir is not None:
            os.makedirs(args.output_dir, exist_ok=True)
    
    accelerator.wait_for_everyone()

    if args.tokenized_ds is None:
        if args.dataset_name is not None:
            # Downloading and loading a dataset from the hub.
            raw_datasets = load_dataset(
                args.dataset_name,
                args.dataset_config_name,
            )
        else:
            data_files = {}
            dataset_args = {}
            if isinstance(args.train_file, dict):
                data_files = args.train_file
            elif args.train_file is not None:
                data_files["train"] = args.train_file
            raw_datasets = load_dataset(
                "json",
                data_files=data_files,
                **dataset_args,
            )
    else:
        raw_datasets = None

    # Load pretrained model and tokenizer
    if args.config_name:
        config = AutoConfig.from_pretrained(args.config_name, trust_remote_code=True)
    elif args.model_name_or_path:
        config = AutoConfig.from_pretrained(args.model_name_or_path, trust_remote_code=True)
    else:
        raise ValueError(
            "You are instantiating a new config instance from scratch. This is not supported by this script."
        )

    if args.tokenizer_name:
        tokenizer = AutoTokenizer.from_pretrained(args.tokenizer_name, use_fast=not args.use_slow_tokenizer, trust_remote_code=True)
    elif args.model_name_or_path:
        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path, use_fast=not args.use_slow_tokenizer, trust_remote_code=True)
    else:
        raise ValueError(
            "You are instantiating a new tokenizer from scratch. This is not supported by this script."
            "You can do it from another script, save it, and load it from here, using --tokenizer_name."
        )

    if args.model_name_or_path:
        if args.use_qlora:
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=torch.bfloat16,
            )
            device_index = accelerator.local_process_index
            device_map = {"": device_index} # force data-parallel training.
            model = AutoModelForCausalLM.from_pretrained(
                args.model_name_or_path,
                from_tf=bool(".ckpt" in args.model_name_or_path),
                config=config,
                load_in_4bit=True,
                quantization_config=bnb_config,
                device_map=device_map,
                torch_dtype=torch.bfloat16,
                trust_remote_code=True
            )
        else:
            model = AutoModelForCausalLM.from_pretrained(
                args.model_name_or_path,
                from_tf=bool(".ckpt" in args.model_name_or_path),
                config=config,
                low_cpu_mem_usage=args.low_cpu_mem_usage,
                trust_remote_code=True
            )
    else:
        logger.info("Training new model from scratch")
        model = AutoModelForCausalLM.from_config(config)


    # no default pad token for llama!
    # here we add all special tokens again, because the default ones are not in the special_tokens_map
    if isinstance(tokenizer, LlamaTokenizer) or isinstance(tokenizer, LlamaTokenizerFast):
        num_added_tokens = tokenizer.add_special_tokens({
            "bos_token": "<s>",
            "eos_token": "</s>",
            "unk_token": "<unk>",
            "pad_token": "<pad>",
        })
        assert num_added_tokens in [0, 1], "LlamaTokenizer should only add one special token - the pad_token, or no tokens if pad token present."
    elif isinstance(tokenizer, GPTNeoXTokenizerFast) or 'falcon' in args.model_name_or_path:
        num_added_tokens = tokenizer.add_special_tokens({
            "pad_token": "<pad>",
        })
        assert num_added_tokens == 1, "GPTNeoXTokenizer should only add one special token - the pad_token."
    elif isinstance(tokenizer, GPT2Tokenizer) and isinstance(model, OPTForCausalLM):
        num_added_tokens = tokenizer.add_special_tokens({'unk_token': '<unk>'})

    # We resize the embeddings only when necessary to avoid index errors. If you are creating a model from scratch
    # on a small vocab and want a smaller embedding size, remove this test.
    embedding_size = model.get_input_embeddings().weight.shape[0]
    if len(tokenizer) > embedding_size:
        model.resize_token_embeddings(len(tokenizer))

    if args.use_lora:
        if args.use_qlora:
            model = prepare_model_for_kbit_training(model, use_gradient_checkpointing=args.gradient_checkpointing)

        logger.info("Initializing LORA model...")
        peft_config = LoraConfig(
            task_type=TaskType.CAUSAL_LM, 
            inference_mode=False, 
            r=args.lora_rank, 
            lora_alpha=args.lora_alpha, 
            lora_dropout=args.lora_dropout,
            target_modules=args.lora_modules,
        )
        model = get_peft_model(model, peft_config)
        # peft breaks flash attention due to casting norms to fp32. This fixes it back up.
        # See https://github.com/huggingface/peft/issues/790
        from llama_flash_attn_monkey_patch import upcast_layer_for_flash_attention
        model = upcast_layer_for_flash_attention(model, torch.bfloat16)
        model.print_trainable_parameters()

    if args.tokenized_ds is None:
        # Preprocessing the datasets.
        if "prompt" in raw_datasets["train"].column_names and "completion" in raw_datasets["train"].column_names:
            encode_function = partial(
                encode_with_prompt_completion_format,
                tokenizer=tokenizer,
                max_seq_length=args.max_seq_length,
            )
        elif "messages" in raw_datasets["train"].column_names:
            encode_function = partial(
                encode_with_messages_format,
                tokenizer=tokenizer,
                max_seq_length=args.max_seq_length,
            )
        else:
            raise ValueError("You need to have either 'prompt'&'completion' or 'messages' in your column names.")
        
        with accelerator.main_process_first():
            lm_datasets = raw_datasets.map(
                encode_function,
                batched=False,
                num_proc=args.preprocessing_num_workers,
                load_from_cache_file=not args.overwrite_cache,
                remove_columns=[name for name in raw_datasets["train"].column_names if name not in ["input_ids", "labels", "attention_mask"]],
                desc="Tokenizing and reformatting instruction data",
            )
            lm_datasets.set_format(type="pt")
            lm_datasets = lm_datasets.filter(lambda example: (example['labels'] != -100).any())


        ds, sample_metadata_map, category_indices = ss_set_train_dataset(lm_datasets, args, accelerator)

        args.tokenized_ds = store_tokenized_ds(args, ds, sample_metadata_map, category_indices)

    ds, sample_metadata_map, category_indices = load_tokenized_dataset(args.tokenized_ds) 
        
    train_dataset = ds['train']
    train_eval_dataset = ds['test'] 

    # Collator does not like strings.
    train_dataset = train_dataset.remove_columns(['sample_uuid', 'category_name'])
    train_eval_dataset = train_eval_dataset.remove_columns(['sample_uuid', 'category_name'])
    

    # Log a few random samples from the training set:
    for index in random.sample(range(len(train_dataset)), 3):
        logger.info(f"Sample {index} of the training set: {train_dataset[index]}.")

    # DataLoaders creation:
    train_dataloader = DataLoader(
        train_dataset, 
        shuffle=True, 
        collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest", pad_to_multiple_of=8),
        batch_size=args.per_device_train_batch_size
    )
    train_eval_dataloader = DataLoader(
        train_eval_dataset, 
        shuffle=False, 
        collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest", pad_to_multiple_of=8),
        batch_size=args.per_device_eval_batch_size
    )

    # Optimizer
    # Split weights in two groups, one with weight decay and the other not.
    no_decay = ["bias", "layer_norm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            "weight_decay": args.weight_decay,
        },
        {
            "params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)],
            "weight_decay": 0.0,
        },
    ]
    if args.use_qlora:
        from bitsandbytes.optim import AdamW
        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=args.learning_rate,
            optim_bits=8 if args.use_8bit_optimizer else 32,
            is_paged=True
        )
    else:
        optimizer = torch.optim.AdamW(optimizer_grouped_parameters, lr=args.learning_rate)

    # Scheduler and math around the number of training steps.
    overrode_max_train_steps = False
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if args.max_train_steps is None:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
        overrode_max_train_steps = True

    # Create the learning rate scheduler.
    # Note: the current accelerator.step() calls the .step() of the real scheduler for the `num_processes` times. This is because they assume 
    # the user initialize the scheduler with the entire training set. In the case of data parallel training, each process only
    # sees a subset (1/num_processes) of the training set. So each time the process needs to update the lr multiple times so that the total 
    # number of updates in the end matches the num_training_steps here.
    # Here we need to set the num_training_steps to either using the entire training set (when epochs is specified) or we need to multiply the 
    # num_training_steps by num_processes so that the total number of updates matches the num_training_steps.
    # num_training_steps_for_scheduler = args.max_train_steps if overrode_max_train_steps else args.max_train_steps * accelerator.num_processes

    lr_scheduler = get_scheduler(
        name=args.lr_scheduler_type,
        optimizer=optimizer,
        num_warmup_steps=int(args.max_train_steps * args.warmup_ratio),
        num_training_steps=args.max_train_steps * args.gradient_accumulation_steps,
    )

    # Prepare everything with `accelerator`.
    model, optimizer, train_dataloader, train_eval_dataloader, lr_scheduler = accelerator.prepare(
        model, optimizer, train_dataloader, train_eval_dataloader, lr_scheduler
    )

    # We need to recalculate our total training steps as the size of the training dataloader may have changed.
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if overrode_max_train_steps:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
    # Afterwards we recalculate our number of training epochs
    args.num_train_epochs = math.ceil(args.max_train_steps / num_update_steps_per_epoch)

    # Figure out how many steps we should save the Accelerator states
    checkpointing_steps = args.checkpointing_steps
    if checkpointing_steps is not None and checkpointing_steps.isdigit():
        checkpointing_steps = int(checkpointing_steps)

    # We need to initialize the trackers we use, and also store our configuration.
    # The trackers initializes automatically on the main process.
    if args.with_tracking:
        experiment_config = vars(args)
        # TensorBoard cannot log Enums, need the raw value
        experiment_config["lr_scheduler_type"] = experiment_config["lr_scheduler_type"].value
        accelerator.init_trackers("open_instruct", experiment_config)

    # Train!
    total_batch_size = args.per_device_train_batch_size * accelerator.num_processes * args.gradient_accumulation_steps

    logger.info("***** Running training *****")
    logger.info(f"  Num examples = {len(train_dataset)}")
    logger.info(f"  Num Epochs = {args.num_train_epochs}")
    logger.info(f"  Instantaneous batch size per device = {args.per_device_train_batch_size}")
    logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
    logger.info(f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}")
    logger.info(f"  Total optimization steps = {args.max_train_steps}")
    # Only show the progress bar once on each machine.
    progress_bar = tqdm(range(args.max_train_steps), disable=not accelerator.is_local_main_process)
    completed_steps = 0
    starting_epoch = 0

    # Potentially load in the weights and states from a previous save
    if args.resume_from_checkpoint:
        if args.resume_from_checkpoint is not None or args.resume_from_checkpoint != "":
            checkpoint_path = args.resume_from_checkpoint
            path = os.path.basename(args.resume_from_checkpoint)
        else:
            # Get the most recent checkpoint
            dirs = [f.name for f in os.scandir(os.getcwd()) if f.is_dir()]
            dirs.sort(key=os.path.getctime)
            path = dirs[
                -1
            ]  # Sorts folders by date modified, most recent checkpoint is the last
            checkpoint_path = path
            path = os.path.basename(checkpoint_path)

        accelerator.print(f"Resumed from checkpoint: {checkpoint_path}")
        accelerator.load_state(path)
        # Extract `epoch_{i}` or `step_{i}`
        training_difference = os.path.splitext(path)[0]

        if "epoch" in training_difference:
            starting_epoch = int(training_difference.replace("epoch_", "")) + 1
            resume_step = None
            completed_steps = starting_epoch * num_update_steps_per_epoch
        else:
            # need to multiply `gradient_accumulation_steps` to reflect real steps
            resume_step = (
                int(training_difference.replace("step_", ""))
                * args.gradient_accumulation_steps
            )
            starting_epoch = resume_step // len(train_dataloader)
            completed_steps = resume_step // args.gradient_accumulation_steps
            resume_step -= starting_epoch * len(train_dataloader)

    # update the progress_bar if load from checkpoint
    progress_bar.update(completed_steps)

    if not args.use_spaced_sampling:
        for epoch in range(starting_epoch, args.num_train_epochs):
            model.train()
            total_loss = 0
            if (
                args.resume_from_checkpoint
                and epoch == starting_epoch
                and resume_step is not None
            ):
                # We skip the first `n` batches in the dataloader when resuming from a checkpoint
                active_dataloader = accelerator.skip_first_batches(
                    train_dataloader, resume_step
                )
            else:
                active_dataloader = train_dataloader
            for step, batch in enumerate(active_dataloader):
                with accelerator.accumulate(model):
                    outputs = model(input_ids=batch['input_ids'], labels=batch['labels'], attention_mask=batch['attention_mask'], use_cache=False)
                    loss = outputs.loss
                    # We keep track of the loss at each logged step
                    total_loss += loss.detach().float()
                    accelerator.backward(loss)
                    # clip gradient norm. don't do this with deepspeed
                    if accelerator.sync_gradients and args.clip_grad_norm > 0:
                        accelerator.clip_grad_norm_(model.parameters(), args.clip_grad_norm)
                    optimizer.step()
                    optimizer.zero_grad()
                    lr_scheduler.step()       

                # Checks if the accelerator has performed an optimization step behind the scenes
                if accelerator.sync_gradients:
                    progress_bar.update(1)
                    completed_steps += 1
                    if args.logging_steps and completed_steps % args.logging_steps == 0:
                        avg_loss = accelerator.gather(total_loss).mean().item() / args.gradient_accumulation_steps / args.logging_steps
                        logger.info(f"  Step: {completed_steps}, LR: {lr_scheduler.get_last_lr()[0]}, Loss: {avg_loss}")
                        if args.with_tracking:
                            accelerator.log(
                                {
                                    "learning_rate": lr_scheduler.get_last_lr()[0],
                                    "train_loss": avg_loss,
                                },
                                step=completed_steps,
                            )
                        total_loss = 0
                        
                    if isinstance(checkpointing_steps, int):
                        if completed_steps % checkpointing_steps == 0:
                            output_dir = f"step_{completed_steps}"
                            if args.output_dir is not None:
                                output_dir = os.path.join(args.output_dir, output_dir)
                            save_with_accelerate(accelerator, model, tokenizer, output_dir, args)

                    if completed_steps >= args.max_train_steps:
                        evaluate_training(args, accelerator, model, tokenizer, train_eval_dataloader, sample_metadata_map, completed_steps)
                        break


            evaluate_training(args, accelerator, model, tokenizer, train_eval_dataloader, sample_metadata_map, completed_steps)
            if args.checkpointing_steps == "epoch":
                output_dir = f"epoch_{epoch}"
                if args.output_dir is not None:
                    output_dir = os.path.join(args.output_dir, output_dir)
                save_with_accelerate(accelerator, model, tokenizer, output_dir, args)
    else:

        from accelerate.utils import gather_object
        ss_checkpointing_steps = args.max_train_steps // args.num_train_epochs // 2
        true_epoch_steps = args.max_train_steps // args.num_train_epochs
        epoch = 0
        done_training = False
        multiprocess_debug = True 

        # SpaceSampling  Config
        log_review = True
        warmup_sample_ratio = 0.1 if args.debug else args.ss_warmup_ratio
        new_samples_ratio = args.ss_new_sample_ratio
        total_sample_cnt = len(train_dataset) 
        category_used_new_sample_cnt = defaultdict(int)
        used_new_sample_cnt = 0
        schedule = defaultdict(list)
        review_logs = []
        reviews = {}
        prev_epoch_selected_samples = None
        num_categories = len(category_indices) 
        min_score = None

        debug_list = {}
        
        while True:
            total_loss = 0

            #############################
            curr_new_samples_ratio = warmup_sample_ratio if epoch == 0 else new_samples_ratio
            new_samples_cnt = int(curr_new_samples_ratio * total_sample_cnt)

            if args.ss_stratified_new_samples:
                pre_category_new_samples_cnt = max(new_samples_cnt // num_categories, 1)
                new_sample_indices = []
                for c, ids in category_indices.items():
                    i = category_used_new_sample_cnt[c]
                    selected_cnt = 0 
                    while i < len(ids) and selected_cnt < pre_category_new_samples_cnt:
                        new_sample_indices.append(ids[i])
                        category_used_new_sample_cnt[c] += 1
                        i += 1
                        selected_cnt += 1
                        
            else:
                new_sample_indices = list(range(used_new_sample_cnt, min(used_new_sample_cnt + new_samples_cnt, total_sample_cnt)))

            used_new_sample_cnt += len(new_sample_indices)

            if epoch == 0:
                # Warmup epoch, no evaluation
                curr_epoch_sample_ids = new_sample_indices 
            else:
                # TODO if schedule_samples and len(schedule_samples) >= int(new_samples_cnt * 0.7):
                # Evaluate to select the sample after warmup epoch
                selected_samples = []
                scheduled_samples = [s[0] for s in schedule[epoch]]
                
                candidate_samples_ids = new_sample_indices + scheduled_samples 
                if not args.ss_skip_prev_samples and prev_epoch_selected_samples is not None:
                    candidate_samples_ids.extend(prev_epoch_selected_samples)
                    
                candidate_samples_ids = gather_object(candidate_samples_ids)
                candidate_samples_ids = list(set(candidate_samples_ids))
                
                if not candidate_samples_ids:
                    # No candidate samples
                    schedule_epochs = list(schedule.keys())
                    if schedule_epochs and max(schedule_epochs) > epoch:
                        # Still samples in the schedule
                        continue
                    else:
                        print('Finished training. nothing in the schedule')
                        done_training = True
                        break

                if multiprocess_debug:
                    print(f"From process={accelerator.process_index} Epoch {epoch} candidate samples: {len(candidate_samples_ids)} - Exp: {candidate_samples_ids[:3]} ... {candidate_samples_ids[-3:]}\n")

                curr_epoch_candidates = train_dataset.select(candidate_samples_ids)
                eval_dataloader = DataLoader(
                    curr_epoch_candidates,
                    shuffle=False,
                    collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest"),
                    batch_size=args.per_device_eval_batch_size,
                )
                eval_dataloader = accelerator.prepare(eval_dataloader)
                
                model.eval()
                all_scores = []
                for _, eval_batch in enumerate(tqdm(eval_dataloader, disable=not accelerator.is_local_main_process, desc="Evaluating")):
                    curr_samples_ids, curr_scores, pred_str_list, label_str_list = evaluate_batch(model, eval_batch, tokenizer)
                    for sample_id, score, pred_str, label_str in zip(curr_samples_ids, curr_scores, pred_str_list, label_str_list):
                        
                        all_scores.append(score)

                        review = reviews.get(sample_id)
                        if review is None: 
                            # First Review
                            first_review = True
                            new_review = CustomSuperMemo2.first_review(score)
                            reviews[sample_id] = new_review
                            interval = new_review.interval
                        else:
                            # Subsequent Reviews
                            first_review = False
                            new_review = review.review(score)
                            interval = new_review.interval
                            reviews[sample_id] = new_review
                            
                        max_target_epoch = max(list(schedule.keys()), default=0)
                        for e in range(epoch, max_target_epoch + 1):
                            if sample_id in schedule[e]:
                                # The sample will be either used or re-scheduled
                                del schedule[e][sample_id]
                                
                                
                        # Droping
                        drop_sample = False
                        drop_reason = "NA"
                        if args.ss_dropout:
                            is_hard_sample = score < 3 and reviews[sample_id].total_repetitions > 2
                            is_easy_sample = reviews[sample_id].successful_repetitions >= 2
                            is_super_easy_sample = first_review and score > 3
                            if is_hard_sample:
                                drop_sample = True
                                drop_reason = "Hard"
                            elif is_easy_sample:
                                drop_sample = True
                                drop_reason = "Easy"
                            elif is_super_easy_sample:
                                drop_sample = True
                                drop_reason = "Super Easy"

                        is_competency_drop = min_score is not None and score < min_score
                        if is_competency_drop:
                            interval = min_score - score

                        if not drop_sample:
                            if interval <= 1: 
                                # Learning should be done now
                                target_epoch = epoch
                                selected_samples.append((sample_id, score))
                            else:
                                # Learning might be done in the future
                                target_epoch = epoch + interval
                                schedule[target_epoch].append((sample_id, score))
                        else:
                            target_epoch = -1

                        if True:
                            if sample_id not in debug_list:
                                debug_list[sample_id] = sample_metadata_map['train'][str(sample_id)]
                            else:
                                if debug_list[sample_id] != sample_metadata_map['train'][str(sample_id)]:
                                    raise ValueError(f"Sample metadata mismatch: {debug_list[sample_id]} != {sample_metadata_map['train'][str(sample_id)]}")
                        
                        if log_review:
                            log_entry = {
                                "Sample_id": sample_id,
                                "Sample_uuid": sample_metadata_map['train'][str(sample_id)]['sample_uuid'],
                                "Sample category": sample_metadata_map['train'][str(sample_id)]['category_name'],
                                "Score": score,
                                "Is first review": first_review,
                                "Review epoch": epoch,
                                "Target Learn epoch": target_epoch,
                                "Interval": interval,
                                "Easiness": reviews[sample_id].easiness,
                                "Successful repetitions": reviews[sample_id].successful_repetitions,
                                "Total repetitions": reviews[sample_id].total_repetitions,
                                "Is doped": drop_sample,
                                "Drop reason": drop_reason,
                                "Prediction": pred_str,
                                "Label": label_str,
                            }
                            review_logs.append(log_entry)
                
                if log_review:
                    write_review_log(args.output_dir, review_logs, epoch, accelerator)
                    review_logs = []
                                
                if len(selected_samples) == 0 and used_new_sample_cnt >= total_sample_cnt:
                    # TODO ?
                    print('No samples selected for learning in this epoch.')
                    break
                            

                # Sort the selected samples by score, start from the lowest score
                sorted_samples = sorted(selected_samples, key=lambda x: x[1])

                curr_epoch_sample_ids = [sample[0] for sample in sorted_samples]

                if not args.ss_skip_prev_samples:
                    # We need to make sure next iteration that we learned the samples properly
                    prev_epoch_selected_samples = curr_epoch_sample_ids.copy() 

                all_scores = gather_object(all_scores)
                competency = np.mean(all_scores)
                if multiprocess_debug:
                    print(f"\nFrom process={accelerator.process_index} Epoch {epoch} Competency: {competency}\n")
                    
                if args.ss_competency_drop:
                    min_score = (5 - int(competency)) - 1
                    
                

                if args.with_tracking:
                    accelerator.log(
                        {
                            "competency": competency, 
                        },
                        step=completed_steps,
                    )
                
            curr_epoch_sample_ids = gather_object(curr_epoch_sample_ids)  
            curr_epoch_sample_ids = list(set(curr_epoch_sample_ids))
            

            if multiprocess_debug:
                print(f"From process={accelerator.process_index} Epoch {epoch} selected samples: {len(curr_epoch_sample_ids)} - Exp: {curr_epoch_sample_ids[:3]} ... {curr_epoch_sample_ids[-3:]}\n")
                
            if len(curr_epoch_sample_ids) == 0:
                print('No samples selected for learning in this epoch.')
                continue

            curr_epoch_samples = train_dataset.select(curr_epoch_sample_ids)

            if True:
                # Spot check if the ids are correct
                for index in random.sample(range(len(curr_epoch_samples)), min(len(curr_epoch_samples), 3)):
                    test_sample_id = curr_epoch_samples[index]['id'].item()
                    if train_dataset[test_sample_id]['id'] != test_sample_id: 
                        raise ValueError(f"Id mismatch: {train_dataset[test_sample_id]['id']} != {test_sample_id}")

            active_dataloader = DataLoader(
                curr_epoch_samples,
                shuffle=True if epoch == 0 else False,
                collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest"),
                batch_size=args.per_device_train_batch_size,
            )
            active_dataloader = accelerator.prepare(active_dataloader)
            #############################
            
            model.train()

            for _, batch in enumerate(tqdm(active_dataloader, disable=not accelerator.is_local_main_process, desc=f"Training: Epoch={epoch}")):
                with accelerator.accumulate(model):
                    outputs = model(input_ids=batch['input_ids'], labels=batch['labels'], attention_mask=batch['attention_mask'], use_cache=False)
                    loss = outputs.loss
                    # We keep track of the loss at each logged step
                    total_loss += loss.detach().float()
                    accelerator.backward(loss)
                    optimizer.step()
                    optimizer.zero_grad()
                    lr_scheduler.step()       

                # # Checks if the accelerator has performed an optimization step behind the scenes
                if accelerator.sync_gradients:
                    progress_bar.update(1)
                    completed_steps += 1

                    if args.logging_steps and completed_steps % args.logging_steps == 0:
                        avg_loss = accelerator.gather(total_loss).mean().item() / args.gradient_accumulation_steps / args.logging_steps
                        logger.info(f"  Step: {completed_steps}, LR: {lr_scheduler.get_last_lr()[0]}, Loss: {avg_loss}")
                        if args.with_tracking:
                            accelerator.log(
                                {
                                    "learning_rate": lr_scheduler.get_last_lr()[0],
                                    "train_loss": avg_loss,
                                },
                                step=completed_steps,
                            )
                        total_loss = 0
                        
                    if completed_steps % ss_checkpointing_steps == 0:
                        output_dir = f"step_{completed_steps}"
                        if args.output_dir is not None:
                            output_dir = os.path.join(args.output_dir, output_dir)
                        save_with_accelerate(accelerator, model, tokenizer, output_dir, args)
                        
                    if completed_steps % true_epoch_steps == 0:
                        evaluate_training(args, accelerator, model, tokenizer, train_eval_dataloader, sample_metadata_map, completed_steps)

                    if completed_steps >= args.max_train_steps:
                        print('Reached max train steps')
                        evaluate_training(args, accelerator, model, tokenizer, train_eval_dataloader, sample_metadata_map, completed_steps)
                        done_training = True
                        break

            # End of epoch
            if False and epoch == 5:
                # Debugging
                break

            if done_training:
                break

            epoch += 1
                

    if args.with_tracking:
        accelerator.end_training()

    if args.output_dir is not None:
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            tokenizer.save_pretrained(args.output_dir)
        save_with_accelerate(accelerator, model, tokenizer, args.output_dir, args)


if __name__ == "__main__":
    main()
